/**
 * Copyright (c) 2024 Mattia Gallacchi
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

#include <string.h>
#include <stdio.h>
#include "pico/stdlib.h"
#include "pico/cyw43_arch.h"
#include "pico/binary_info.h"
#include "hardware/i2c.h"

#include "lwip/pbuf.h"
#include "lwip/udp.h"

#include "sht30.h"

#define WIFI_SSID               "IGIB_WiFi"
#define WIFI_PASSWORD           "1234567890"
#define UDP_PORT                4444
#define BEACON_MSG_LEN_MAX      127
#define BEACON_TARGET           "255.255.255.255"
#define BEACON_INTERVAL_MS      2500
#define CONNECT_TIMEOUT         10000 // 10s

enum ERROR_CODES {
    NONE=0,
    WIFI_CONNECT,
    SHT30_CONNECT,
    TIMER_SETUP,
    UDP_LOOP,
};

bool flash_led_callback(struct repeating_timer *t)
{
    static bool led = true;

    // Flash the led to show activity
    cyw43_arch_gpio_put(CYW43_WL_GPIO_LED_PIN, led);
    led = !led;

    return true;
}

bool run_udp_beacon(SHT30 *sensor)
{
    bool error = false, led = false;
    ip_addr_t addr;
    SHT30::measure mes = {0, 0};

    struct udp_pcb* pcb = udp_new();
    ipaddr_aton(BEACON_TARGET, &addr);

    while (!error) {
        struct pbuf *p = pbuf_alloc(PBUF_TRANSPORT, BEACON_MSG_LEN_MAX+1, PBUF_RAM);
        char *req = (char *)p->payload;
        memset(req, 0, BEACON_MSG_LEN_MAX+1);

        if (!sensor->single_measure(SHT30::HIGH, mes)) {
            error = true;
        }   

        snprintf(req, BEACON_MSG_LEN_MAX, "%04X:%0.2f;%04X:%0.2f\n",mes.raw_temperature, mes.temperature, mes.raw_humidity, mes.humidity);
        err_t er = udp_sendto(pcb, p, &addr, UDP_PORT);
        pbuf_free(p);
        if (er != ERR_OK) {
            printf("Failed to send UDP packet! error=%d", er);
        } else {
            printf("Sent packet %f:%f\n", mes.temperature, mes.humidity);
        }

        // Note in practice for this simple UDP transmitter,
        // the end result for both background and poll is the same

#if PICO_CYW43_ARCH_POLL
        // if you are using pico_cyw43_arch_poll, then you must poll periodically from your
        // main loop (not from a timer) to check for Wi-Fi driver or lwIP work that needs to be done.
        cyw43_arch_poll();
        sleep_ms(BEACON_INTERVAL_MS);
#else
        // if you are not using pico_cyw43_arch_poll, then WiFI driver and lwIP work
        // is done via interrupt in the background. This sleep is just an example of some (blocking)
        // work you might be doing.
        sleep_ms(BEACON_INTERVAL_MS);
#endif
    }

    return error;
}

void led_errors(ERROR_CODES code)
{
    uint32_t interval = 250;
    uint32_t pause = 2000;
    uint8_t flash_count = (uint8_t)code;

    cyw43_arch_gpio_put(CYW43_WL_GPIO_LED_PIN, 0);

    while(1) {
        for(uint8_t i = 0; i < flash_count; i++) {
            cyw43_arch_gpio_put(CYW43_WL_GPIO_LED_PIN, 1);
            sleep_ms(interval);
            cyw43_arch_gpio_put(CYW43_WL_GPIO_LED_PIN, 0);
            sleep_ms(interval);
        }

        sleep_ms(pause);
    }
}

int main()
{
    struct repeating_timer *timer = new struct repeating_timer;
    SHT30 *sht;
    ERROR_CODES err = NONE;

    stdio_init_all();

    /** WiFi init **/
    printf("\n*** Init WiFi ***\n");    
    if (cyw43_arch_init()) {
        while(1) {
            printf("Failed to initialise WiFi\n");
            sleep_ms(1000);
        }
    }

    cyw43_arch_enable_sta_mode();

    if(!add_repeating_timer_ms(500, flash_led_callback, NULL, timer)) {
        printf("Error: Failed to add timer\n");
        err = TIMER_SETUP;
        goto end;
    }

    printf("Connecting to Wi-Fi...\n");
    if (cyw43_arch_wifi_connect_timeout_ms(WIFI_SSID, WIFI_PASSWORD, CYW43_AUTH_WPA2_AES_PSK, CONNECT_TIMEOUT)) {
        printf("Error: Failed to connect to %s.\n", WIFI_SSID);
        err = WIFI_CONNECT;
        goto end;
    }
    printf("Connected to %s.\n", WIFI_SSID);

    /**
     * Init SHT30 sensor
    */
    printf("\n*** Init SHT30 sensor ***\n");
    sht = new SHT30();

    if (!sht->is_connected()) {
        printf("Error: Failed to connect SHT30\n");
        err = SHT30_CONNECT;
        delete sht;
        goto end;
    }
    printf("SHT30 connected\n");

    if(!run_udp_beacon(sht)) {
        printf("Error: Failed to start UDP loop");
        err = UDP_LOOP;
        goto end;
    }

    printf("An error occurred stopping\n");

    uint16_t status_reg;
    if (sht->read_status_reg(status_reg))
        printf("Status reg: 0x%04X\n", status_reg);

    end:
        if (timer->alarm_id > 0)
            cancel_repeating_timer(timer);
        led_errors(err);
        return 1;
}